package models

import (
	"encoding/json"
	"log"
	"net/http"

	"forge.grandlyon.com/web-et-numerique/factory/llle_project/backoffice-server/internal/common"
	"forge.grandlyon.com/web-et-numerique/factory/llle_project/backoffice-server/internal/constants"
)

type MonthlyNews struct {
	Year    int    `json:"year"`
	Month   int    `json:"month"`
	Title   string `json:"title"`
	Content string `json:"content"`
}

func (dh *DataHandler) GetAllMonthlyNews(w http.ResponseWriter, r *http.Request) {
	var monthlyNews []MonthlyNews
	dh.sqlClient.Find(&monthlyNews)

	w.Header().Set(constants.ContentType, constants.Json)
	json.NewEncoder(w).Encode(monthlyNews)
	log.Printf("| get all monthlyNews | %v", r.RemoteAddr)
}

func (dh *DataHandler) GetSingleMonthlyNews(w http.ResponseWriter, r *http.Request) {
	year, month, err := common.YearMonthFromRequest(r)
	if err != nil {
		http.Error(w, err.Error(), http.StatusBadRequest)
		return
	}

	w.Header().Set(constants.ContentType, constants.Json)

	var monthlyNews MonthlyNews
	err = dh.sqlClient.Where("year = ? AND month = ?", year, month).First(&monthlyNews).Error
	if err != nil {
		// If not found, answer "not found"
		w.WriteHeader(http.StatusNotFound)
		return
	}
	json.NewEncoder(w).Encode(monthlyNews)
}

func (dh *DataHandler) SaveMonthlyNews(w http.ResponseWriter, r *http.Request) {
	if r.Body == http.NoBody {
		http.Error(w, "request body is empty", http.StatusBadRequest)
		return
	}

	decoder := json.NewDecoder(r.Body)
	var monthlyNews MonthlyNews
	err := decoder.Decode(&monthlyNews)
	if err != nil {
		w.WriteHeader(http.StatusInternalServerError)
		return
	}

	// Check if this monthlyNews exists
	err = dh.sqlClient.Where("year = ? AND month = ?", monthlyNews.Year, monthlyNews.Month).First(&MonthlyNews{}).Error

	// Default title
	if monthlyNews.Title == "" && monthlyNews.Content != "" {
		monthlyNews.Title = "Les nouveautés du service"
	}

	if err != nil {

		// Create a monthlyNews
		err = dh.sqlClient.Create(&MonthlyNews{Year: monthlyNews.Year, Month: monthlyNews.Month, Title: monthlyNews.Title, Content: monthlyNews.Content}).Error
		if err != nil {
			w.WriteHeader(http.StatusInternalServerError)
			return
		}
		w.WriteHeader(http.StatusCreated)
		json.NewEncoder(w).Encode(monthlyNews)

		log.Printf("| new monthlyNews | year : %d month : %d title : %v| %v", monthlyNews.Year, monthlyNews.Month, monthlyNews.Title, r.RemoteAddr)
		return

	} else {
		// Update title
		err = dh.sqlClient.Model(&MonthlyNews{}).Where("year = ? AND month = ?", monthlyNews.Year, monthlyNews.Month).Update("title", monthlyNews.Title).Error
		if err != nil {
			w.WriteHeader(http.StatusInternalServerError)
			return
		}
		// Update content
		err = dh.sqlClient.Model(&MonthlyNews{}).Where("year = ? AND month = ?", monthlyNews.Year, monthlyNews.Month).Update("content", monthlyNews.Content).Error
		if err != nil {
			w.WriteHeader(http.StatusInternalServerError)
			return
		}

		json.NewEncoder(w).Encode(monthlyNews)
		log.Printf("| updated monthlyNews | year : %d month : %d title : %v| %v", monthlyNews.Year, monthlyNews.Month, monthlyNews.Title, r.RemoteAddr)
	}
}

func (dh *DataHandler) DeleteMonthlyNews(w http.ResponseWriter, r *http.Request) {
	year, month, err := common.YearMonthFromRequest(r)
	if err != nil {
		http.Error(w, err.Error(), http.StatusBadRequest)
		return
	}

	err = dh.sqlClient.Where("year = ? AND month = ?", year, month).Delete(&MonthlyNews{}).Error
	if err != nil {
		http.Error(w, err.Error(), http.StatusNotFound)
		return
	}

	w.WriteHeader(http.StatusOK)
	w.Write([]byte("successful delete"))
}
